@@ -1,6 +1,7 @@
 import json
 import logging
 import sqlite3
+from abc import abstractmethod
 from collections.abc import MutableMapping
 from enum import Enum, unique
 from typing import Any, Dict, List, Union
@@ -36,16 +37,18 @@ def flatten(d, parent_key="", sep="_"):
 
 
 class Expression():
+    @abstractmethod
     def compile(self, expression: str):
         """Compiles an expression
 
         Args:
             expression (str): expression
         """
-        pass
+        raise NotImplementedError
 
-    def search(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
-        """Executes the expression on a given data
+    @abstractmethod
+    def search(self, data: Dict[str, Any]) -> Dict[str, Any]:
+        """Executes the expression on a single data entry
 
         Args:
             data (List[Dict[str, Any]]): Data
@@ -53,41 +56,52 @@ class Expression():
         Returns:
             List[Dict[str, Any]]: Transformed data
         """
-        pass
+        raise NotImplementedError
 
-
-class SQLExpression(Expression):
-    def compile(self, expression: str):
-        # we turn off check_same_thread to gain performance benefit by reusing the same connection object. safe to use since we are only creating in memory structures
-        self.conn = sqlite3.connect(":memory:", check_same_thread=False)
-        self.expression = expression
-
-    def filter(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
-        """Tests a where clause for an SQL statement
+    @abstractmethod
+    def search_bulk(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
+        """Executes the expression in bulk on a list of data entries
 
         Args:
             data (List[Dict[str, Any]]): Data
 
+        Returns:
+            List[Dict[str, Any]]: Transformed data
+        """
+        raise NotImplementedError
+
+    def filter(self, data: List[Dict[str, Any]], tombstone: bool = False) -> List[Union[Dict[str, Any], None]]:
+        """Tests a where clause for an SQL statement
+
+        Args:
+            data (List[Dict[str, Any]]): Data
+            tombestone: if True, returns None for filtered values
+
         Returns:
             List[Dict[str, Any]]: Filtered data
         """
-        data_inner = flatten_data(data)
-        cte_clause = self._get_cte(data_inner)
+        # filter is essentially using a single field returning 0 or 1 for the condition
+        if tombstone:
+            return [row[0] if row[1] else None for row in zip(data,self.search_bulk(data))]
+        else:
+            return [row[0] for row in zip(data,self.search_bulk(data)) if row[1]]
 
-        column_names = data_inner[0].keys()
-
-        # fetch the CTE and bind the variables
-        data_values = [row.get(column_name) for row in data_inner for column_name in column_names]
-        self.conn.row_factory = sqlite3.Row
-
-        cursor = self.conn.execute(
-            f"{cte_clause} select * from data where {self.expression}", data_values
-        )
-        return [dict(row) for row in cursor.fetchall()]
+class SQLExpression(Expression):
+    def compile(self, expression: str):
+        # we turn off `check_same_thread` to gain performance benefit by reusing the same connection object
+        # safe to use since we are only creating in memory structures
+        self.conn = sqlite3.connect(":memory:", check_same_thread=False)
+        # we support both single field expressions and multiple fields
+        self._is_single_field = True
+        try:
+            self._fields = json.loads(expression)
+            self._is_single_field = False
+        except json.JSONDecodeError:
+            # this is not a json, treat as a simple expression
+            self._fields = {"expr": expression}
 
     def _get_cte(self, data: List[Any]) -> str:
         # builds a CTE expression for fetching in memory data
-
         column_names = data[0].keys()
         columns_clause = ','.join(f"`{col}`" for col in column_names)
 
@@ -98,22 +112,23 @@ class SQLExpression(Expression):
         # use a CTE to create the in memory data structure
         return f"with data({columns_clause}) as (values {values_clause})"
 
-    def search(self, data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
-        try:
-            fields = json.loads(self.expression)
-            new_data = {}
-            for field in fields:
-                new_data[field] = self.exec_sql(data, fields[field])
-            return new_data
-        except json.JSONDecodeError:
-            # this is not a json, treat as a simple expression
-            return self.exec_sql(data, self.expression)
+    def search_bulk(self, data: List[Dict[str, Any]]) -> Any:
+        results = self.exec_sql(data, self._fields)
+        if self._is_single_field:
+            # treat as a simple expression
+            return [x.get("expr") for x in results]
+        else:
+            return results
 
-    def exec_sql(self, data: List[Dict[str, Any]], expression: str) -> List[Dict[str, Any]]:
+    def search(self, data: Dict[str, Any]) -> Any:
+        return self.search_bulk([data])[0]
+
+    def exec_sql(self, data: List[Dict[str, Any]], expressions: Dict[str, str]) -> List[Dict[str, Any]]:
         """Executes an SQL statement
 
         Args:
             data (List[Dict[str, Any]]): Data
+            expressions: Dict[str, str]: Expressions
 
         Returns:
             List[Dict[str, Any]]: Query result
@@ -126,13 +141,14 @@ class SQLExpression(Expression):
 
         # fetch the CTE and bind the variables
         data_values = [row.get(colname) for row in data_inner for colname in column_names]
-        self.conn.row_factory = sqlite3.Row
-        logger.debug(f"{cte_clause} select {expression} from data)")
-        cursor = self.conn.execute(
-            f"{cte_clause} select {expression} from data", data_values
-        )
 
-        return cursor.fetchone()[0]
+        # expressions clause
+        expressions_clause = ", ".join([f"{expression} as `{column_name}`" for column_name,expression in expressions.items()])
+        self.conn.row_factory = sqlite3.Row
+        statement = f"{cte_clause} select {expressions_clause} from data"
+        logger.debug(statement)
+        cursor = self.conn.execute(statement, data_values)
+        return [dict(x) for x in cursor.fetchall()]
 
 
 class JMESPathExpression(Expression):
@@ -141,14 +157,13 @@ class JMESPathExpression(Expression):
 
     def compile(self, expression: str):
         self.expression = jmespath.compile(expression)
-        self.filter_expression = jmespath.compile(f"[?{expression}]")
 
-    def filter(self, data: Union[Dict[str, Any], List[Dict[str, Any]]]) -> Any:
-        return self.filter_expression.search(data, options=self.options)
-
-    def search(self, data: Union[Dict[str, Any], List[Dict[str, Any]]]) -> Any:
+    def search(self, data: Dict[str, Any]) -> Any:
         return self.expression.search(data, options=self.options)
 
+    def search_bulk(self, data: List[Dict[str, Any]]) -> Any:
+        return [self.search(row) for row in data]
+
 
 def compile(language: Language, expression: str) -> Expression:
     """Gets a compiled expression class based on the language
@@ -164,6 +179,8 @@ def compile(language: Language, expression: str) -> Expression:
         expression_class = JMESPathExpression()
     elif language == Language.SQL.value:
         expression_class = SQLExpression()
+    else:
+        raise ValueError(f"unknown expression language {language}")
 
     expression_class.compile(expression)
     return expression_class
